﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Xml.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Text.RegularExpressions;
using Ionic.Zip;
using DataFlowConverter;

namespace CodeRetrieverDataflow
{
    public partial class Form1 : Form
    {
        /*
         * This file holds all of the written methods used by the application.
        */

        /*
         * This sets up for the selected profile when a new profile has been chosen.
         *
         * Called From:
         *   ChangeProfile()
         *   tag_list_SelectedIndexChanged()
         *
         * Inputs:
         *
         *   profile: The name of the current profile.
         *   
         *   profileLabel: The label used to indicate the current profile being used. Ex: "Individual".
         *   
         *   tagList: The ListBox object which holds the profile tags, the main functional groups for a 
         *              profile. Ex: "Spouse".
         *   
         *   subTagList: The ListBox object which holds the profile sub tags. Ex: "Name".
         *   
         *   subTagCombList: The ListBox object which holds the profile sub tag combinations. Ex: "First".
         *   
         *   manipList: The ListBox object which holds the various manipulations for a particular profile
         *              list item, like for SSN, a manipulation would be "999999999".
         *              
         *   condList: The ListBox object which holds Ruby pseudo code conditional expressions for certain
         *              profile tags. Ex: for "Sole Proprietor" a conditional expression would be "if 
         *              (business name) end".
         *              
         *   formBoxListBox: The ListBox object which holds the boxes associated with a document. Used in
         *                   renumbering and truncating selections.
         *                   
         *   srcArray: The array that holds the list items used in populating the tagList.
        */
        public void SetupProfile(string profile, Label profileLabel, ListBox tagList, ListBox subTagList, ListBox subTagCombList, ListBox manipList, ListBox condList, ListBox formBoxListBox, string[] srcArray)
        {
            // Clear (empty) the tag lists [ListBoxes]:
            tagList.Items.Clear();
            subTagList.Items.Clear();
            subTagCombList.Items.Clear();
            manipList.Items.Clear();
            condList.Items.Clear();

            /*
             * These are cleared by DisplayMnemonicList():
             *  mnemonicListBox
             *  renumList
            */

            /*
             * Clear selections in the ListBox formBoxListBox, and Force the ListBox to scroll back to 
             * the top of the list. Currently not doing this.
            */
            //formBoxListBox.ClearSelected();
            //formBoxListBox.TopIndex = 0;

            // Display the name of the selected  profile to the user if applicable.
            // It is not required when we are leaving common block processing.
            if (!String.IsNullOrEmpty(profile))
            {
                profileLabel.Text = string.Format("{0} Profile", profile);
                tssLabelStatus.Text = profileLabel.Text;
            }

            // Set the Tag List per selected profile:
            foreach (string str in srcArray) { tagList.Items.Add(str); }
        }

        /*
        * This uses the OpenFileDialog class in allowing the user to select a file
        * which will be used in obtaining box information pertaining to a document.
        *
        * Called From:
        *   rubySourceCodeFiledocumentBoxNumbersToolStripMenuItem_Click()
        *   xMLFiledocumentBoxNumbersAndBoxWidthsToolStripMenuItem_Click()
        *
        * Inputs:
        *   dir: The initial file path from which a file may be selected.
        *   type: The type of file we will read to extract the boxes from, either "RBW" or "XML".
        *
        * Returns the full path and file name of the selected file or null if no file was selected.
        */
        public string GetFileName(string dir, string type)
        {
            string fileName = null;
            OpenFileDialog ofd = new OpenFileDialog();

            ofd.InitialDirectory = dir;

            // We completely control the parameter 'type' so if not 'XML' we default to Ruby source files.
            //ofd.Filter = (type == "XML" ? "XML files|*.xml" : "Ruby source files|*.rbw");
            //ofd.Filter = (type != "Ruby" ? "ZIP file|*.zip|XML file|*.xml" : "Ruby source file|*.rbw");

            ///////////////////////////////////////////////////////////////////////
            // This is where the Read option from the top menu lets the user select 
            // the form .zip and builds the box list.
            ///////////////////////////////////////////////////////////////////////
            
            ofd.Filter = (type != "Ruby" ? "ZIP file|*.zip|XML file|*.xml" : "Ruby source file|*.rbw");

            System.Windows.Forms.DialogResult dr = ofd.ShowDialog();

            if (dr == DialogResult.OK)
            {
                // This holds the path and the file name with extension.
                fileName = ofd.FileName;
            }
            return fileName;
        }

        /*
        * This extracts the "DOC.XML" file from the ZIP file identified by the argument zipFileName.
        *
        * Called From:
        *   xMLFiledocumentBoxNumbersAndBoxWidthsToolStripMenuItem_Click()
        *
        * Inputs:
        *   zipFileName: The full path and file name of the zip file selected by the user.
        *
        * Returns the full path and file name of the extracted "DOC.XML" file or null if the exraction
        *   process did not work out well.
        */
        public string ExtractXmlFileFromZipFile(string zipFileName)
        {
            string docXmlFile = null;
            string tmpPath;
            string password = "5e$ACre7ru";

            try
            {
                // We first need to get the temp path where we will store the extracted XML file and we will
                // delete the file after reading it for box information.
                tmpPath = Path.GetTempPath();

                if (!String.IsNullOrEmpty(tmpPath))
                {
                    using (ZipFile zip = ZipFile.Read(zipFileName))
                    {
                        foreach (ZipEntry e in zip)
                        {
                            if (e.FileName.ToLower() == "doc.xml")
                            {
                                if (e.UsesEncryption)
                                {
                                    e.Password = password;
                                }
                                e.Extract(tmpPath, ExtractExistingFileAction.OverwriteSilently);

                                // Set the path and file name of the extracted XML file to be returned to caller.
                                docXmlFile = tmpPath + "doc.xml";
                            }
                        }
                    }
                }
            }
            catch (System.Security.SecurityException ex)
            {
                MessageBox.Show(ex.ToString(), "Could not get the path to the Temp folder.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
            return docXmlFile;
        }

        /*
        * This builds a list of FormBox objects from a supplied .RBW file. The list
        * comes from the ***** Field Data ***** section in the ruby file.
        *
        * The maximum number of characters that the box can hold is NOT stored
        * as this information comes from the XML file "NO XML" is stored instead.
        * ALL box numbers from a form are stored.
        *
        * Called From:
        *   rubySourceCodeFiledocumentBoxNumbersToolStripMenuItem_Click()
        *
        * Inputs:
        *   fileName: The .RBW file that contains the boxes.
        *
        * Returns a list of all extracted FormBox objects.
        */
        public List<FormBox> BuildBoxArrayRbw(string fileName)
        {
            List<FormBox> tmpList = new List<FormBox>();
            string line;

            try
            {
                using (StreamReader sr = new StreamReader(fileName))
                {
                    while (((line = sr.ReadLine()) != null) && (!Regex.IsMatch(line, @"\*+\s*Calculations\s*\*+")))
                    {
                        Match match = Regex.Match(line, @"^box(\d+_\d+|box\d+)");
                        if (match.Success)
                        {
                            tmpList.Add(new FormBox(match.Value, "NO XML"));
                        }
                    }
                }
            }
            catch (IOException ex)
            {
                MessageBox.Show(ex.ToString(), "File could not be read.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

            return tmpList;
        }

        /*
        # This builds a list of FormBox objects from a supplied .XML file, starting
        # with the first box, ending with the last box. Along with the box number,
        # the maximum number of characters that the box can hold is also stored if the
        # font style is 'correct' else we use "MAN" to indicate to the user that a manual
        # entry of the width must be performed. ALL box numbers for a form are stored.
        #
        # Called From: xMLFiledocumentBoxNumbersAndBoxWidthsToolStripMenuItem_Click()
        #
        # Inputs:
        #
        #   fileName: The .XML file that contains the boxes.
        #
        # Returns a list of all extracted FormBox objects.
        */
        public List<FormBox> BuildBoxArrayXml(string fileName)
        {
            List<FormBox> tmpList = new List<FormBox>();
            string line;
            string box = null;
            string width = null;
            float coordDiff = 0.0f;
            float size = 0.0f;
            Match match;
            bool notFirstPass = false;

            try
            {
                using (StreamReader sr = new StreamReader(fileName))
                {
                    while ((line = sr.ReadLine()) != null)
                    {
                        // Remove whitespace:
                        line = Regex.Replace(line, @"\s+", "");

                        // Looking for the starting line of a box:
                        match = Regex.Match(line, @"ref=""(box\d+_\d+|box\d+)""");
                                               
                        if (match.Success)
                        {
                            /*
                            If this is not the first box, we must finalize the processing for the current box first.
                            */
                            if (notFirstPass)
                            {
                                /*
                                "MAN" means we do not have the correct font type for this box to calculate
                                the maximum width automatically, the user will need to set it manually.
                                */
                                if (width == null)
                                {
                                    width = "MAN";
                                }
                                tmpList.Add(new FormBox(box, width));

                                // Reset for the next box processing:
                                width = null;
                                coordDiff = 0.0f;
                                size = 0.0f;
                            }
                            else
                            {
                                notFirstPass = true;
                            }

                            // Grab the box value from the match. Groups are indexed from 1, not 0, index 1 is our box.
                            box = match.Groups[1].Value;
                        }

                        /*
                        # Now look for the associated coordinates and determine the maximum number of characters that
                        # the box can hold. This number is used for truncation purposes. We can only determine this number
                        # when the font is either "Courier New" or "OCRA A Extended". When the font is one of these types,
                        # we check if there is a font size indicated for the box. If there is no indication, we default to size = 12.
                        # These two fonts have the same character size, so an "i" has the same width as a "W" for example.
                        #
                        # The width is calculated as: (right * 10.0 - left * 10.0) / (size * 6). The width will be set to "MAN" to
                        # flag a Manual truncation for truncating purposes when the font type is not "Courier New" or "OCRA A Extended".
                        */
                        match = Regex.Match(line, @"<bboxleft=""(\d+\.\d+)""top=""\d+\.\d+""right=""(\d+\.\d+)""bottom=""\d+\.\d+""\/>$");

                        if (match.Success)
                        {
                            // Index 2 is the right coord and index 1 is the left coord, we do not need top and bottom coords.
                            // We have the box number, now we have the left and right coordinates, we next need to know the font type and size.
                            // Had to use Math.Round() here to catch slight precision goofs, like getting a value of 71.99994 instead of 72,
                            // or 72.00003 instead of 72.
                            coordDiff = (float)Math.Round(Convert.ToSingle(match.Groups[2].Value) * 10.0f - Convert.ToSingle(match.Groups[1].Value) * 10.0f);
                        }

                        // Looking for font information - if we find the types mentioned above we can set the width:
                        if (Regex.IsMatch(line, @"<font"))
                        {
                            if (Regex.IsMatch(line, @"OCRAExtended|CourierNew"))
                            {
                                // We can calculate the maximum width. We default to 12 points if no font size info.
                                match = Regex.Match(line, @"size=""(\d+)""");
                                size = (match.Success ? Convert.ToSingle(match.Groups[1].Value) : 12.0f);

                                // We were using the Ruby .to_i method for this portion in the TK version, which truncates a float to an
                                // integer: i.e. 1.5.to_i -> 1. This makes sense as we are determing how many characters fit in a box, no
                                // such thing as a partial character. Being extra cautious here with 'size'.
                                if (size > 0.0f)
                                {
                                    width = Math.Floor(coordDiff / (size * 6.0f)).ToString("F0");
                                }
                            }
                        }
                    }
                }

                if (box != null)
                {
                    /*
                    We need to close the processing for the last box.
                
                    "MAN" means we do not have the correct font type for this box to calculate
                    the maximum width automatically, the user will need to set it manually.
                    */
                    if (width == null)
                    {
                        width = "MAN";
                    }
                    tmpList.Add(new FormBox(box, width));
                }
            }
            catch (IOException ex)
            {
                MessageBox.Show(ex.ToString(), "File could not be read.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (OverflowException ex)
            {
                MessageBox.Show(ex.ToString(), "Overflow Exception while converting to a single precision floating point number.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

            return tmpList;
        }

        /*
         * This routine searches through the global buffer xmlDocument, looking for the dataflow code
         * that pertains to the supplied mnemonic variable.
         * 
         * If the search was successful, a dfObject is returned which contains the boxes associated
         * with the dataflow code, the description of each box [aids in renumbering], the mnemonic string,
         * and the actual dataflow code - either Ruby or JS depending on the arg getJS - see below.
         * 
         * An error message is displayed if the search ws not successful, with null being returned.
         * 
         * Called From:
         * 
         *  globalRenumButton_Click()
         *  insertMnemonicButton_Click()
         *  retrieveButton_Click()
         *  
         * Inputs:
         * 
         *  mnemonic: A string that describes the code snippet to retrieve. e.g.
         *      "IND|INDIVIDUAL:NAME::FIRST"
         *  getJS: 'True' for retrieving JS code, else we will retrieve Ruby code.
         *  
         * Returns:
         * 
         *  dfObject: An initialized DataFlow object if we found the code snippet.
         *  null: If we did not find the code snippet.
        */
        public DataFlow GetDataflowByMnemonic(String mnemonic, bool getJS)
        {
            int foundData = 0;

            dfObject = new DataFlow(mnemonic);

            // START: DOING THIS JUST TO OUTPUT THE STARTING AND STOPPING MNEMONICS FOR EQUIFAX.

            //retrieveText.AppendText(string.Format("</{0}>\n", mnemonic));
            //retrieveText.AppendText(string.Format("<{0}>\n\n", mnemonic));
            //return dfObject;

            // STOP: DOING THIS JUST TO OUTPUT THE STARTING AND STOPPING MNEMONICS FOR EQUIFAX.

            if (xmlDocument != null)
            {
                IEnumerable<XElement> nodes;
                IEnumerable<XElement> subNodes;
                XElement node = null;

                try
                {
                    nodes = xmlDocument.Descendants().Where(e => ((e.Name == "Block") && e.HasAttributes && (e.Attribute("mnemonicId") != null) && (e.Attribute("mnemonicId").Value == mnemonic)));

                    if (nodes.Count() > 0)
                    {
                        foundData += 1;

                        node = nodes.First();

                        subNodes = node.Elements().Where(e => ((e.Name == "Boxes") || (e.Name == "Desc") || (e.Name == (getJS ? "JavaScriptCode" : "RubyCode"))));

                        foreach (XElement e in subNodes)
                        {
                            if (e.Name == "Boxes")
                            {
                                dfObject.BoxNum = e.Value.Split(',').ToList();
                                dfObject.NewBoxNum = e.Value.Split(',').ToList();

                                foreach (string box in dfObject.BoxNum)
                                {
                                    dfObject.BoxWidth.Add("");
                                }

                                foundData += 1;
                            }
                            else if (e.Name == "Desc")
                            {
                                dfObject.BoxDesc = e.Value.Split('|').ToList();
                                foundData += 1;
                            }
                            else if ((e.Name == "RubyCode") || (e.Name == "JavaScriptCode"))
                            {
                                dfObject.DataFlowCode = e.Value;
                                foundData += 1;
                            }
                            else
                            {
                                // Should not happen once we build the XML file.
                                MessageBox.Show(string.Format("Unknown node detected for mnemonic:\n\n{0}\n\nNode Name: {1}", mnemonic, e.Name), "Unknown Node Detected", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            }
                        }
                    }
                }
                catch (Exception e)
                {
                    MessageBox.Show(e.ToString());
                }
            }
            else
            {
                MessageBox.Show("xmlDocument is null.", "GetDataflowByMnemonic(): No XML Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            if (foundData == 4)
            {
                /*
                 * Special code for the Equifax Profile State List Page:
                 * 
                 *      If the mnemonic starts with "EQU STATE LIST|##:" [quotes not included], we will
                 *      replace the "##" with the selected state in the tagList.
                 *      
                 *      We will do this for the mnemonic, the dataflow code and the description.
                */
                if (Regex.Match(mnemonic, @"^EQU STATE LIST\|##:").Success)
                {
                    dfObject.Mnemonic = mnemonic.Replace("##", profileTag);
                    dfObject.DataFlowCode = dfObject.DataFlowCode.Replace("##", profileTag);

                    for (int i = 0; i < dfObject.BoxDesc.Count; i++)
                    {
                        dfObject.BoxDesc[i] = dfObject.BoxDesc[i].Replace("##", profileTag);
                    }
                }

                return dfObject;
            }
            else
            {
                /*
                 * If the mnemonic starts with "EQU STATE LIST|##" [quotes not included], we will replace
                 * the "##" with the selected state in the tagList. Doing this so users do not get confused.
                */
                ProcessErrors("Error!", 4, (Regex.Match(mnemonic, @"^EQU STATE LIST\|##").Success ? mnemonic.Replace("##", profileTag) : mnemonic), null);

                return null;
            }
        }

        /*
         * This routine retrieves from the XML file/document JS 'special' code. This is a section
         * in the XML file separate from the 'regular' data flow code block snippets.
         * 
         * An error message is displayed if the search was not successful, with null being returned.
         * 
         * Called From:
         * 
         *  generateProfileButton_Click()
         *  GetCond()
         *  
         * Inputs:
         * 
         *  specialStr: A string that identifies the desired special code.
         *  
         * Returns:
         * 
         *  A string containing the JS special code. Returns null if the JS special code
         *  could not be retrieved from the XML document.
        */
        public string GetJSSpecialCode(String specialStr)
        {
            int foundData = 0;
            string jsCode = null;

            if (xmlDocument == null)
            {
                IEnumerable<XElement> nodes;
                IEnumerable<XElement> subNodes;
                XElement node = null;

                try
                {
                    nodes = xmlDocument.Descendants().Where(e => ((e.Name == "Block") && e.HasAttributes && (e.Attribute("mnemonicId") != null) && (e.Attribute("mnemonicId").Value == specialStr)));

                    if (nodes.Count() > 0)
                    {
                        foundData += 1;
                        node = nodes.First();
                        subNodes = node.Elements().Where(e => (e.Name == "JavaScriptCode"));

                        foreach (XElement e in subNodes)
                        {
                            foundData += 1;
                            jsCode = e.Value + Environment.NewLine;
                        }
                    }
                }
                catch (Exception e)
                {
                    MessageBox.Show(e.ToString());
                }
            }
            else
            {
                MessageBox.Show("xmlDocument is null.", "GetJSSpecialCode(): No XML Data", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            if (foundData == 2)
            {
                return jsCode;
            }
            else
            {
                ProcessErrors("Error!", 5, specialStr, null);
                return null;
            }
        }

        /*
        #
        # This routine displays error/warning messages to the user via MessageBox.Show().
        #
        # Called From:
        #
        #   GetJSSpecialCode()
        #   GetDataflowByMnemonic()         
        #   globalRenumButton_Click()
        #   renumModifyButton_Click()
        #   truncAddButton_Click()
        #   insertMnemonicButton_Click()
        #   retrieveButton_Click()
        #   GetCond()
        #   DoCommonBlocks()
        #
        # Inputs:
        #
        #   errType: Either "Error" or "Warning".
        #   errNum: The error number which is our index into the string array errMsg.
        #   var1 and var2: If not NULL, we insert the value of 'var' into our error message.
        */
        public void ProcessErrors(string errType, int errNum, string var1, string var2)
        {
            string[] errMsg = { 
		        "  You must first select a Profile.\n",
		        "  Your selected text must not lie inside other text.\n\n  Operation aborted.\n",
		        string.Format("  {0}\n\n  Is already in the Selection List.\n", var1),
		        "  Selected Selection List elements must be in one block with no gaps.\n",
		        string.Format("  {0}\n\n  Bad selection - Operation aborted.\n", var1),
		        string.Format("  Could not retrieve JS code for:\n\n  {0}\n\n  Operation aborted.\n", var1),
		        "  You must select the same number of elements in each list.\n",
		        "  You may select only one Selection List element at a time.\n",
		        string.Format("  You must enter a Box Offset when you are retrieving for more than 1\n  {0}.\n\n  Operation aborted.\n", var1),
		        string.Format("  Box Offset will be ignored while retrieving for 1 {0}.\n\n  Box Offset only makes sense when you are retrieving for more than 1\n  {1}.\n", var1, var1),
		        "  You must select 1 box from the Dataflow Code area and 1 box from\n  the Box List.\n\n  Operation aborted.\n",
		        "  This Conditional Expression has an 'else' portion that inserts 'box1'\n  into your retrieved code.\n\n  You can renumber this box by using the Renumber Box button or by\n  manually typing in the correct box number yourself.\n",
		        "  You must first select text in the Dataflow Code window.\n\n  Operation aborted.\n",
                string.Format("  \"{0}\" is not a valid {1} value.\n\n  Please enter at least 1 and at most 3 numeric digits.\n", var1, var2),
                string.Format("  Unexpected null returned from:\n\n  {0}\n\n  Operation aborted.\n", var1),
		        "  You must first select one item from the Conditional Dataflow list.\n\n  Operation aborted.\n",
                string.Format("  Unexpected null/empty returned from:\n\n  {0}\n\n  Operation aborted.\n", var1),
                string.Format("  {0} - Duplicate Profile Declaration Error:\n\n  {1}\n  Is already present in Dataflow Code window.\n\n  Operation aborted.\n", var1, var2),
                string.Format("  Generate $ Variables Error:\n\n  {0}  Is already present in Dataflow Code window.\n\n  Operation aborted.\n", var1),
                "  Settings Property Default Mode Error:\n\n  The Settings Property contains a bad mode value.\n\n  The Retriever will default to 'Retrieve Ruby Code'.\n"
                };

            MessageBoxIcon icon;

            // We have complete control over the arguments to this method, so we default to MessageBoxIcon.Warning.
            icon = (errType == "Error" ? MessageBoxIcon.Error : MessageBoxIcon.Warning);
            MessageBox.Show(errMsg[errNum], errType, MessageBoxButtons.OK, (errType == "Error" ? MessageBoxIcon.Error : MessageBoxIcon.Warning));
        }

        /*
        # This method performs the following:
        #
        # 1) Re-synch the global dfObjectList with the 'Selection List' to keep them 1:1.
        #
        # inserting will be true if we are inserting into the 'Selection List'. Deleting
        # from the 'Selection List' will have this set to false.
        #
        # Called From:
        #   insertMnemonicButton_Click()
        #   deleteMnemonicButton_Click()
        #
        # Inputs:
        #
        #   dfObject: An object of DataFlow. This holds all info pertaining to the dataflow snippet we are retrieving.
        #               This object holds the retrieved data flow, box information, the mnemonic which describes the
        #               dataflow snippet we are retrieving, and other pertinent information.
        #
        #   mnemonicListBox: A ListBox that holds the mnemonics that are displayed in the 'Selection List'.
        #
        #   dfObjectList: A list of DataFlow objects. This list holds the dfObject(s) associated with the 'Selection List'.
        #               We renumber the data flow code [during retrieval] using the objects in this list - see class definition
        #               for DataFlow.
        #
        #   inserting: A boolean to indicate if we are inserting or deleting to/from the 'Selection List'.
        #
        # Returns:
        #   A synchronized dfObjectList.
        */
        public List<DataFlow> SynchDfObjList(DataFlow dfObject, ListBox mnemonicListBox, List<DataFlow> dfObjectList, bool inserting)
        {
	        List<String> tmpList = new List<String>();
	        List<DataFlow> dfObjectTmpList = new List<DataFlow>();
            DataFlow delObj;
            bool found;

	        /*
	        This deletes the dfObject if it is already in the global dfObjectList. The reason this is being done is to force
	        the capture of the .NewBoxNum List data. This is a problem when inserting over is allowed. So we catch it here.
	        Note that this is for inserting into the 'Selection List', Deletion(s) from the list will not cause this problem, so
	        we first check the 'inserting' argument. 
	        
	        What is the 'problem'? We are matching on mnemonic only, so inserting over a list element with the same mnemonic
	        but different box prefix elements would result in the old box prefix values being retained as we would see the
	        associated object already present in the dfObjectList when in fact that would be a good thing, however we want
	        to pick-up any box prefix changes.
	        */
            if (inserting)
            {
                delObj = null;

                for (int index = 0; index < dfObjectList.Count; index++)
                {
                    if (dfObjectList[index].Mnemonic == dfObject.Mnemonic)
                    {
				        delObj = dfObjectList[index];
				        break;
                    }
                }

                if (delObj != null)
                {
    		        dfObjectList.Remove(delObj);
                }
            }

            for (int index = 0; index < mnemonicListBox.Items.Count; index++)
            {
                // We split each item based on " || ", which gives us a two element array per split. The 
                // second array element is the mnemonic. We control the strings in the mnemonicListBox,
                // so we always know there is a " || " component per list item.
                tmpList.Add(mnemonicListBox.Items[index].ToString().Split(new string[] { " || " }, StringSplitOptions.None)[1]);
            }

            for (int index = 0; index < tmpList.Count; index++)
            {
                found = false;

                for (int i = 0; i < dfObjectList.Count; i++)
                {
                    if (dfObjectList[i].Mnemonic == tmpList[index])
                    {
                        dfObjectTmpList.Add(dfObjectList[i]);
                        found = true;
                        break;
                    }
                }

                if (!found)
                {
                    dfObjectTmpList.Add(dfObject);
                }

            }

            return dfObjectTmpList;
        }

        /*
         * This method renumbers and/or truncates the boxes [if applicable] present in the DataFlowCode
         * member in each of the dfObject(s) which are passed in via dfObjList, which has been setup to
         * contain only those boxes that will be modified.
         * 
         * Called From:
         *      RetrieveDataflow()
         *      
         * Inputs:
         *      dfObjList: A list of DataFlow objects. This list holds the dfObject(s) which we may be
         *                  renumbering and/or truncating the boxes present in the DataFlowCode member(s).
         *                  
         * Returns:
         *      A string array of renumbered and/or truncated dataflow code - just the actual dataflow
         *      code, no dfObject(s).
        */
        public List<string> ModifyDf(List<DataFlow> dfObjList)
        {
	        List<string> strList = new List<string>();
            String box;
            String var;
	        String str;
            String lineCopy;
	        String truncStr;
            int indx;
            int width = 0;
            Match match;
            Boolean retJS = (retMode == retrieverMode.retrieveJS);

	        /*
             * Using .upcase to prevent boxes being renumbered twice. For example, box_num ['box1_1',
             * 'box1_2'] and new_box_num ['box1_2', 'box1_3'] would cause trouble without the upcase
             * combined with gsub. box1_1 would be changd to box1_2, however, the next array position
             * would result in the newly numbered box1_2 being again renumbered to box1_3 as per the
             * new_box_num array.
	        */
            try
            {
                for (int index = 0; index < dfObjList.Count; index++)
                {
                    // First we do the renumbering:
                    str = dfObjList[index].DataFlowCode;

                    for (int i = 0; i < dfObjList[index].BoxNum.Count; i++)
                    {
                        str = str.Replace(dfObjList[index].BoxNum[i], dfObjList[index].NewBoxNum[i].ToUpper());
                    }

                    str = str.Replace("BOX", "box");

                    /*
                     * Now do the truncation if applicable [note that all arrays are the same size by
                     * definition]:
                    */
                    truncStr = "";

                    /*
                    We need to employ StringSplitOptions.RemoveEmptyEntries because, for example, "ABC\n"
                    split using StringSplitOptions.None would yield a 2 element array, with one element
                    being an empty string, which causes one more iteration thru the loop [does not hurt
                    anything text wise], but causes an extra newline to be re-attached down below since we
                    are splitting on newline chars.
                    */
                    string[] linesArray = str.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.RemoveEmptyEntries);

                    foreach (string line in linesArray)
                    {
                        // We can not change the foreach loop var 'line', so we make a copy.
                        lineCopy = line;

                        // We would not be here if we were converting Ruby to JS, so.....:
                        if (retJS)
                        {
                            // The JS way [converter style] to print a value to a box...
                            match = Regex.Match(line, @"(\s*SFForms.StorageManager.setValue\(formInstanceId\s*,\s*[""'])(box\d+_\d+|box\d+)([""']\s*,\s*)(.+)");
                        }
                        else
                        {
                            match = Regex.Match(lineCopy, @"(\s*print +')(box\d+_\d+|box\d+)(', +)(.+)");
                        }

                        if (match.Success)
                        {
                            box = match.Groups[2].Value;
                            var = match.Groups[4].Value;

                            indx = dfObjList[index].NewBoxNum.IndexOf(box);
                            int.TryParse(dfObjList[index].BoxWidth[indx], out width);

                            if (width > 0)
                            {
                                /*
                                If we have a match, $' is the string after the match.
                                If the string does not end in an ' if...' then var is already 
                                equal to what we want. i.e. "print 'box1_10', tmp" would give us var
                                equal to "tmp", while "print 'box1_10', tmp if !tmp.empty?" would
                                at first have var equal to "tmp if !tmp.empty?", that is where $'
                                comes in to give us var equal to "tmp".
                            
                                How does it work again? The following Regex.Replace() call takes 'var'
                                as the input string and checks if the string ENDS with the pattern
                                @" +if.+$", in our discussion, that would be " if !tmp.empty?" for 'var'
                                equal to "tmp if !tmp.empty?". If the string 'var' does end with that pattern,
                                we REPLACE THE PATTERN MATCH with the string contents AFTER the match
                                [the substitution parameter "$'"], BUT, in the case of a match there is NO
                                remaining characters after the match since we match on a pattern that runs up
                                to the end of the string, so we are left with the contents of the string UP TO
                                the match.
                            
                                This truncation logic behaves exactly like the Ruby version, it is up to the
                                user to perform sensible truncation. For a Ruby example, if the 'line' looks
                                like <print 'box1_1', "#{$1} #{$2}">, I added the '<' and '>', and they decided
                                to truncate this Dataflow, and the calculated width was say, 10, then 'line'
                                would look like: <print 'box1_1', "#{$1} #{$2}"[0,10]>.
                                */
                                var = Regex.Replace(var, (retJS ? @"\s*\);$" : @" +if.+$"), "$'");

                                /*
                                Now we add the truncation part. We must first use Regex.Escape(var) so that the
                                characters in 'var' are interpreted literally, since we may have, for example,
                                'var' equal to "#{$1}" where  $ and { have special meaning in a regular expression.
                                */
                                Regex regex = new Regex(Regex.Escape(var));

                                /*
                                Here we substitute the FIRST [from the left] instance of 'var' with the contents
                                of 'var' immediately followed by the truncation characters. The 1 indicates only
                                one replacement operation.
                                */
                                if (retJS)
                                {
                                    lineCopy = regex.Replace(line, String.Format("({0}).substr(0,{1})", var, width), 1);
                                }
                                else
                                {
                                    lineCopy = regex.Replace(line, String.Format("{0}[0,{1}]", var, width), 1);
                                }
                            }
                        }

                        // Re-attach our stripped newline char.
                        truncStr += (lineCopy + "\n");
                    }

                    strList.Add(truncStr);
                }
            }
            catch (ArgumentOutOfRangeException ex)
            {
                MessageBox.Show(ex.ToString(), "Bad index value encountered in ModifyDF().", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (ArgumentNullException ex)
            {
                MessageBox.Show(ex.ToString(), "A null 'old value' String.Replace Argument was encountered in ModifyDF().", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (ArgumentException ex)
            {
                MessageBox.Show(ex.ToString(), "An empty 'old value' String.Replace Argument was encountered in ModifyDF().", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

            return strList;
        }

        /*
         * This method determines which conditional expression we are going to process, using the
         * supplied arguments and returns an initialized CondExp object.
         * 
         * Called From:
         *    RetrieveDataflow()
         *    condButton_Click()
         *
         * Inputs:
         * 
         *    profileTag: This is the currently selected tag from the tagList (e.g. "Spouse" for
         *      the individual profile).
         *      
         *    condList: A ListBox which holds the various conditional expression strings.
         *    
         *
         * Returns:
         *     A string from the condList based on profileTag.
        */
        public CondExp GetCond(string profileTag, ListBox condList)
        {
            CondExp condExp = null;

            if (retMode == retrieverMode.retrieveJS)
            {
                string specialStr = null;

                specialStr = GetJSSpecialCode(condList.SelectedItem.ToString());

                if (!String.IsNullOrEmpty(specialStr))
                {
                    // Throwing in an extra newline to match the Ruby retrieval.
                    condExp = new CondExp(specialStr + "\n", null);
                }
            }
            else if (retMode == retrieverMode.retrieveRuby)
            {
                /*
                 * Keeping this approach as it has been for several years:
                 * 
                 * We need to determine what the conditional expression is: The conditional expression
                 * is either an 'if (..)' or an 'if (..) else' expression. We are currently
                 * handling just the 'if (..)'. The 'end' portion will be addressed by the calling
                 * routine.
                 * 
                 * Also - the conditional list elements are worded so that a new person/someone not
                 * proficient in Ruby can read and understand, so we cannot simply pull the text and
                 * use that, we need to know the usage context and convert to Ruby here.
                */
                switch (profileTag)
                {
                    case ("SPOUSE"):

                        switch (condList.SelectedIndices[0])
                        {
                            case (0):

                                condExp = new CondExp("if (!individual.married_joint.strip.empty?)\n\n", null);
                                break;

                            case (1):

                                condExp = new CondExp("if (!individual.married_separate.strip.empty?)\n\n", null);
                                break;

                            case (2):

                                condExp = new CondExp("if (!(individual.married_joint + individual.married_separate).strip.empty?)\n\n", null);
                                break;
                        }
                        break;

                    case ("SOLE PROPRIETOR"):

                        switch (condList.SelectedIndices[0])
                        {

                            case (0):

                                condExp = new CondExp("if (!individual.business_name.strip.empty?)\n\n", null);
                                break;
                        }
                        break;
                }
            }

            return condExp;
        }

        /*
        * This method handles common block Dataflow code retrieval in it's entirety as it is a twist from the 'other' retrievals.
        * Common block means dataflow areas on a document that are identical such as Individual Dependent dataflow where each block has
        * the same fields, the only difference is each field is numbered per block. So the Individual profile contains
        * dataflow fields for up to 6 dependents, so the profile has individual.dependent1_ssn up to individual.dependent6_ssn.
        *
        * Called From:
        *   retrieveButton_Click()
        *
        * Inputs:
        *
        *   retrieveStr: The string that contains the retrieved dataflow code thus far.
        *
        *   numBlocks: The number of common blocks [think dependents] we are procesing. It is the same
        *   dataflow code per block, with the exception of the box numbers, which it is assumed, will
        *   differ by a common offset per dataflow item.
        *
        *   noExtraLine: A boolean which indicates if an extra newline char is required.
        *
        *   doCond: A boolean which indicates if conditional processing will be performed.
        *
        *   boxOffsetValue: The common offset between block dataflow box numbers. For example, dependent
        *   1 has a first name in box1_1 and dependent 2 has a first name in box1_6, an offset of 5
        *   exists between all dependent dataflow box numbers.
        *
        *   dfStrList: A list of strings where each list entry is the dataflow code for each mnemonic to
        *   be retrieved.
        *   
        *   blockType: Indicates the type of common block we are processing.
        */
        public string DoCommonBlocks(string retrieveStr, int numBlocks, bool noExtraLine, bool doCond, int boxOffsetValue, List<string> dfStrList, string blockType)
        {
            List<string> blockStrList = new List<string>();
            string[] linesArray;
            string blockStr = "";
            int cnt;
            int lastCnt;

            // Here we run thru the List of dataflow strings, one element in the list is the dataflow code for one mnemonic.
            for (int index = 0; index < dfStrList.Count; index++)
            {
                if (doCond)
                {
                    linesArray = dfStrList[index].Split(new string[] { "\r\n", "\n" }, StringSplitOptions.RemoveEmptyEntries);

                    foreach (string line in linesArray)
                    {
                        blockStrList.Add(string.Format("\t{0}\n", line));
                    }
                }
                else
                {
                    blockStrList.Add(dfStrList[index]);
                }

                blockStrList.Add("\n");
            }

            /*
            These are used in determining if we are going to insert a newline char ["\n"] when we are processing the very
            last element in depStrList - see above loop through the dfStrList.
            */
            cnt = 0;
            lastCnt = numBlocks * blockStrList.Count;

            /*
            We now just run through the blockStrList for each block, renumbering accordingly with no file access.
            boxOffsetValue is used for offsetting box numbers between loop iterations.
            */
            for (int loopNum = 1; loopNum <= numBlocks; loopNum++)
            {
                for (int index = 0; index < blockStrList.Count; index++)
                {

                    cnt += 1;

                    if (cnt == lastCnt)
                    {
                        if ((blockStrList[index] == "\n") && noExtraLine)
                        {
                            break;
                        }
                    }

                    /*
                     * We will first replace the block number holder '#' with the acual block number per
                     * our current loop iteration:
                    */
                    if (blockType == "dependent")
                    {
                        blockStr = blockStrList[index].Replace("dependent#", string.Format("dependent{0}", loopNum));
                    }
                    else
                    {
                        blockStr = blockStrList[index].Replace("officer#", string.Format("officer{0}", loopNum));
                    }

                    /*
                     * For each box pattern we encounter, we will pass the match value to our Delegate
                     * method NewBoxNum(), which will add the offset value to the box number portion of
                     * the mtch and the matched pattern will be replace by the new box pattern with the
                     * adjusted box number.
                    */
                    Regex reg = new Regex(@"box\d+_\d+|box\d+");
                    blockStr = reg.Replace(blockStr, match => NewBoxNum(match, boxOffsetValue * (loopNum - 1)));

                    retrieveStr += blockStr;
                }
            }
            return retrieveStr;
        }

        /*
         * This method handles common block Dataflow box renumbering. This method serves as a Delegate
         * method for reg.Replace.
         * 
         * Called From:
         *    DoCommonBlocks
         *    
         * Inputs:
         *    match: The match on the box pattern, e.g. "box1_5"
         *    
         *    offset: The value to add to the box number portion of the match in order to get the new
         *              box number. For example,
         *              "box1_5" gives 5 as the box number, and if the offset was 3 we would return
         *              "box1_8".
         * Returns:
         *    The renumbered box, "box1_8" using the above example.
        */
        public string NewBoxNum(Match match, int offset)
        {
            int num;
            string[] boxParts = new string[2];

            /*
             * We know we have a pattern such as "box1_17", so we can split on the "_" and parse the
             * second value to get an integer value.
            */
            boxParts = match.Value.Split('_');
            Int32.TryParse(boxParts[1], out num);
            return string.Format("{0}_{1}", boxParts[0], num + offset);
        }

        /*
         * This method counts the number of newlines in a string.
         * 
         * Called From:
         *    GoodSelection()
         *    retrieveButton_Click()
         *    condButton_Click()
         *
         * Inputs:
         *    str: The string we will examine for newline chars.
         *    
         * Returns:
         *    The number of newlines found in argument str.
        */
        //public int CountNewLines(string str)
        public int CountNewLines(ref string str)
        {
            int count = 0;
            int position = 0;

            while ((position = str.IndexOf('\n', position)) != -1)
            {
                count++;

                // Keep moving forward...
                position++;
            }
            return count;
        }

        /*
         * We are entering common block processing for a particular profile. We store the block's
         * 'Parent' 'Selection List' contents for future use and set the global dfObjectList to the 
         * 'Child' blocks copy of dfObjectList and display in the 'Selection List.
         *
         * Called From:
         *   tag_list_SelectedIndexChanged()
         *
         * Inputs:
         *   blockName: The 'Child' block we are entering.
        */
        public void EnteringCommonBlockProcessing(string blockName)
        {
            /*
            # Entering Common Block processing: Dependents, Officers, etc..
            #
            # Because there can be multiple blocks selected, with each having the identical
            # data flow code, we will re-display the Profile Tag list with common block
            # specific elements which the user can click on. From this list they can choose how
            # many blocks to retrieve code for, or <Return> to the parent Profile Tag
            # List. 
            #
            # In this switch-over to common block processing, we will also save the current contents of
            # the Selection List [store a copy of dfObjectList into the mnemonicList] and re-populate
            # this list with the current contents of the block's {block prefix}MnemonicList.
            */
            string[] blockTagArray = { };
            string[] blockSubTagArray = { };
            string blockOffsetNum = "";
            int numBlock = 1;

            switch (blockName)
            {
                case ("IND|DEPENDENTS"):

                    indivMnemonicList = dfObjectList;
                    blockTagArray = indivDepProcArray;
                    blockSubTagArray = indivDepArray;
                    numBlock = numDependents;
                    blockOffsetNum = indivDepOffset;
                    break;

                case ("EQU SUCCESSOR|OFFICERS"):

                    equSuccMnemonicList = dfObjectList;
                    blockTagArray = equOfficerTagArray;
                    blockSubTagArray = equOfficerSubTagArray;
                    numBlock = numEquSuccOfficers;
                    blockOffsetNum = equSuccOfficOffset;
                    break;

                case ("EQU PREDECESSOR|OFFICERS"):

                    equPredMnemonicList = dfObjectList;
                    blockTagArray = equOfficerTagArray;
                    blockSubTagArray = equOfficerSubTagArray;
                    numBlock = numEquPredOfficers;
                    blockOffsetNum = equPredOfficOffset;
                    break;
            }

            tagList.Items.Clear();
            foreach (string str in blockTagArray) { tagList.Items.Add(str); }

            // This is cleared by the calling routine.
            foreach (string str in blockSubTagArray) { subTagList.Items.Add(str); }

            // Activate the 'Box Offset' entry panel and set the offset value:
            boxOffsetPanel.Enabled = true;
            boxOffsetEntry.Text = blockOffsetNum;

            // Clear selections in the ListBox form_box_list_box, and Force the ListBox to scroll back to the top of the list. Currently not doing.
            //formBoxListBox.ClearSelected();
            //formBoxListBox.TopIndex = 0;

            tagListLocked = true;
            tagList.SelectedIndex = numBlock - 1;
        }

        /*
         * We are leaving common block processing for a particular profile. We store the block's
         * Mnemonic List ['Selection List'] and the box offset value for future use.
         *
         * Called From:
         *   tag_list_SelectedIndexChanged()
         *   ChangeProfile()
         *
         * Inputs:
         *   blockName: The 'Child' block we are leaving.
        */
        public void LeavingCommonBlockProcessing(string blockName)
        {
            /*
             * Leaving Common Block processing: Dependents, Officers, etc..
             * 
             * Make a copy of the Selection List so the user can return later and resume common block
             * processing.
            */
            switch (blockName)
            {
                case ("IND|DEPENDENTS"):

                    indivDepMnemonicList = dfObjectList;
                    indivDepOffset = boxOffsetEntry.Text;
                    break;

                case ("EQU SUCCESSOR|OFFICERS"):

                    equSuccOffMnemonicList = dfObjectList;
                    equSuccOfficOffset = boxOffsetEntry.Text;
                    break;

                case ("EQU PREDECESSOR|OFFICERS"):

                    equPredOffMnemonicList = dfObjectList;
                    equPredOfficOffset = boxOffsetEntry.Text;
                    break;
            }

            tagListLocked = false;

            // Clear and De-activate the 'Box Offset' entry field:
            boxOffsetEntry.Clear();
            boxOffsetPanel.Enabled = false;

            // De-activate the 'Truncate' entry field:
            renumTruncPanel.Enabled = false;

            // Put the cursor back in the retrieve_text widget because the Box Offset entry
            // widget may have had it if they used it.
            retrieveText.Select();
        }

        /*
         * This displays the Profile mnemonic list contents. These are mnemonics that have been retrieved
         * and stored as a batch retrieval. Users can build up dataflow snippets and renumber/truncate
         * them and store them in their own list for retrieval as in one batch, or retrieve based on
         * selection.
         * 
         * Globals Affected:
         *  dfObjectList: Set to the mnemonic list we are switching to.
         *  dfObject: Set to null.
         *  renumList: Cleared.
         *  mnemonicListBox.Items: Cleared and then populated with mnemonic list items we are switching to.
         * 
         * Called From:
         *   tag_list_SelectedIndexChanged()
         *   ChangeProfile()
         *
         * Inputs:
         *   listId: The mnemonic list to display.
        */
        public void DisplayMnemonicList(string listId)
        {
            string prefix;
            string box;
            string width;
            int num;

            switch (listId)
            {
                case ("equifax"):
                    dfObjectList = new List<DataFlow>();
                    break;

                case ("equGen"):
                    dfObjectList = equGenMnemonicList;
                    break;

                case ("equSucc"):
                    dfObjectList = equSuccMnemonicList;
                    break;

                case ("equSuccOff"):
                    dfObjectList = equSuccOffMnemonicList;
                    break;

                case ("equPred"):
                    dfObjectList = equPredMnemonicList;
                    break;

                case ("equPredOff"):
                    dfObjectList = equPredOffMnemonicList;
                    break;

                case ("equState"):
                    dfObjectList = equStateMnemonicList;
                    break;

                case ("entity"):
                    dfObjectList = entityMnemonicList;
                    break;

                case ("individual"):
                    dfObjectList = indivMnemonicList;
                    break;

                case ("indivDep"):
                    dfObjectList = indivDepMnemonicList;
                    break;

                case ("preparer"):
                    dfObjectList = prepMnemonicList;
                    break;

                case ("fiduciary"):
                    dfObjectList = fidMnemonicList;
                    break;
            }

            renumList.Items.Clear();
            mnemonicListBox.Items.Clear();
            dfObject = null;

            /*
             * Now convert all df_object(s) in dfObjectList to the format for the Selection List -
             * 'box prefix + mnemonic':
            */
            foreach (DataFlow obj in dfObjectList)
            {
                prefix = "";

                /*
                 * By design, the lists NewBoxNum and BoxWidth are the same size, so we can use a common
                 * index.
                */
                for (int i = 0; i < obj.NewBoxNum.Count; i++)
                {
                    box = obj.NewBoxNum[i];
                    width = (Int32.TryParse(obj.BoxWidth[i], out num) ? (num > 0 ? string.Format("[{0}]", num) : "") : "");

                    prefix += string.Format("{0}{1} ", box, width);
                }

                mnemonicListBox.Items.Add(string.Format("{0}|| {1}", prefix, obj.Mnemonic));
            }
        }

        /*
         * This handles the switch over to a different profile.
         * 
         * Called From:
         *   individualToolStripMenuItem_Click()
         *   entityToolStripMenuItem_Click()
         *   fiduciaryToolStripMenuItem_Click()
         *   preparerToolStripMenuItem_Click()
         *   equifaxToolStripMenuItem_Click()
         *
         * Inputs:
         *   newProfile: The new profile we are switching to.
        */
        public void ChangeProfile(string newProfile)
        {
            string[] profileArray = { };

            /*
             * First determine which profile state we are in currently by looking at the global var
             * 'mnemonic'. We need to save a copy of dfObjectList of the profile state we are switching
             * from so we can use the mnemonic selections when/if we return to that profile.
             * LeavingCommonBlockProcessing() does this for profiles that have common block processing.
            */
            if (!String.IsNullOrEmpty(mnemonic))
            {
                if (Regex.IsMatch(mnemonic, @"^IND\|DEPENDENTS"))
                {
                    LeavingCommonBlockProcessing("IND|DEPENDENTS");
                }
                else if (Regex.IsMatch(mnemonic, @"^IND\|"))
                {
                    indivMnemonicList = dfObjectList;
                }
                else if (Regex.IsMatch(mnemonic, @"^EQU SUCCESSOR\|OFFICERS"))
                {
                    LeavingCommonBlockProcessing("EQU SUCCESSOR|OFFICERS");
                }
                else if (Regex.IsMatch(mnemonic, @"^EQU SUCCESSOR\|"))
                {
                    equSuccMnemonicList = dfObjectList;
                }
                else if (Regex.IsMatch(mnemonic, @"^EQU PREDECESSOR\|OFFICERS"))
                {
                    LeavingCommonBlockProcessing("EQU PREDECESSOR|OFFICERS");
                }
                else if (Regex.IsMatch(mnemonic, @"^EQU PREDECESSOR\|"))
                {
                    equPredMnemonicList = dfObjectList;
                }
                else if (Regex.IsMatch(mnemonic, @"^EQU GENERAL\|"))
                {
                    equGenMnemonicList = dfObjectList;
                }
                else if (Regex.IsMatch(mnemonic, @"^EQU STATE LIST\|"))
                {
                    equStateMnemonicList = dfObjectList;
                }
                else if (Regex.IsMatch(mnemonic, @"^ENT\|"))
                {
                    entityMnemonicList = dfObjectList;
                }
                else if (Regex.IsMatch(mnemonic, @"^FID\|"))
                {
                    fidMnemonicList = dfObjectList;
                }
                else if (Regex.IsMatch(mnemonic, @"^PRE\|"))
                {
                    prepMnemonicList = dfObjectList;
                }
            }

            // Now we deal with the profile state we are switching to:
            switch (newProfile)
            {
                case ("Entity"):

                    mnemonic = "ENT|";
                    DisplayMnemonicList("entity");
                    profileArray = entProfileArray;
                    break;

                case ("Equifax"):

                    DisplayMnemonicList("equifax");
                    mnemonic = "EQU|";
                    profileArray = equProfileArray;
                    break;

                case ("Fiduciary"):

                    mnemonic = "FID|";
                    DisplayMnemonicList("fiduciary");
                    profileArray = fidProfileArray;
                    break;

                case ("Individual"):

                    mnemonic = "IND|";
                    DisplayMnemonicList("individual");
                    profileArray = indivProfileArray;
                    break;

                case ("Preparer"):

                    mnemonic = "PRE|";
                    DisplayMnemonicList("preparer");
                    profileArray = preProfileArray;
                    break;
            }



            /*
             * Clear the tag lists, set the profile label text box and display the appropriate Tag List
             * contents.
            */
            SetupProfile(newProfile, profileLabel, tagList, subTagList, subTagCombList, manipList, condList, formBoxListBox, profileArray);
            profile = newProfile;

            // Setting focus.
            retrieveText.Select();
        }

        /*
         * This stores the Dataflow XML snippet file contents in memory.
         * 
         * Called From:
         *   Form1_Load()
         *
         * Inputs:
         *   xmlDataFile: The XML snippet file.
        */
        public void SetXmlDocument(string xmlDataFile)
        {
            try
            {
                byte[] xml;
                xmlDocument = null;

                FileStream fs = new FileStream(xmlDataFile, FileMode.Open, FileAccess.Read);

                xml = new byte[fs.Length];
                fs.Read(xml, 0, (int)fs.Length);
                fs.Close();

                /*
                 * Preserving new lines in the code as we split on newlines and discard
                 * empty lines in the process during conversion/retrieval. Doing the whole
                 * file rather than regexing inside the CDATA elements, simpler.
                 */
                var str = System.Text.Encoding.Default.GetString(xml);
                str = Regex.Replace(str, @"(\n|\r\n){2,}", "\n \n");
                xml = System.Text.Encoding.Default.GetBytes(str);

                // Initialize the XDocument memory stream.
                MemoryStream memstream = new MemoryStream();
                memstream.Write(xml, 0, xml.Length);
                memstream.Seek(0, SeekOrigin.Begin);
                xmlDocument = XDocument.Load(memstream);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), string.Format("XML read error - cannot load {0} into memory.", xmlDataFile), MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * Switching from 'Converting Ruby to JS' to retrieving either Ruby or Javascript, so restore
         * the GUI interface to it's fullest as required for the retrieving mode.
         * 
         * Called From:
         *   ProcessMode()
        */
        public void RestoreGuiConvToJS()
        {
            tableLayoutPanel2.Enabled = true;
            tableLayoutPanel3.Enabled = true;
            panel13.Enabled = true;
            readToolStripMenuItem.Enabled = true;
            profileToolStripMenuItem.Enabled = true;
            condButton.Enabled = true;
            renumBoxButton.Enabled = true;
            generateProfileButton.Enabled = true;
        }

        /*
         * Switching to 'Converting Ruby to JS', so disabling GUI components that are not
         * applicable and setting applicable button/label text.
         * 
         * Called From:
         *   ProcessMode()
         *   Form1_Load()
        */
        public void DisableGuiConvToJS()
        {
            tableLayoutPanel2.Enabled = false;
            tableLayoutPanel3.Enabled = false;
            panel13.Enabled = false;
            readToolStripMenuItem.Enabled = false;
            profileToolStripMenuItem.Enabled = false;
            condButton.Enabled = false;
            renumBoxButton.Enabled = false;
            generateProfileButton.Enabled = false;

            generateGlobalsButton.Text = "Gen Variables: ON";
            modeLbl.Text = "Convert Ruby To JavaScript";
            retrieveButton.Text = "Convert Ruby To JS";
        }


        


        /*
         * Handles the switch over from one Retriever mode state to another. The Retriever operates
         * in 3 modes: 'Retrieve Ruby Code'; 'Retrieve JS Code'; and 'Convert Ruby Code to JS Code'.
         * 
         * We need to ensure we 'clear out' any pre-existing code so people do not end up utilizing
         * say, Ruby code when they want to utilize JS code.
         * 
         * Called From:
         * 
         *      convertRubyToJSToolStripMenuItem_Click()
         *      retrieveRubyCodeToolStripMenuItem_Click()
         *      retrieveJSCodeToolStripMenuItem_Click()
        */
        public void ProcessMode(retrieverMode toMode)
        {
            switch (toMode)
            {
                case (retrieverMode.retrieveRuby):

                    retrieveText.Clear();

                    switch (retMode)
                    {
                        case (retrieverMode.retrieveRuby):

                            /*
                             * This will be used to signal a reset when starting a new document
                             * but staying in the same mode and we do not want the previous document's
                             * dfObjects to 'get in the way' as the 'Selection List' would still show
                             * the selections from the old document.
                            */
                            ModeResetDataFlow();
                            return;

                        case (retrieverMode.retrieveJS):

                            ModeResetDataFlow();
                            prevRetrieveMode = retrieverMode.retrieveRuby;
                            ModeSetRetRuby();
                            break;

                        case (retrieverMode.convertRubyToJS):

                            // Restore interface since the conversion mode disables it:
                            RestoreGuiConvToJS();

                            if (prevRetrieveMode == retrieverMode.retrieveJS)
                            {
                                ModeResetDataFlow();
                                prevRetrieveMode = retrieverMode.retrieveRuby;
                            }

                            ModeSetRetRuby();
                            break;
                    }

                    break;

                case (retrieverMode.retrieveJS):

                    switch (retMode)
                    {
                        case (retrieverMode.retrieveJS):

                            /*
                             * This will be used to signal a reset when starting a new document
                             * but staying in the same mode and we do not want the previous document's
                             * dfObjects to 'get in the way' as the 'Selection List' would still show
                             * the selections from the old document.
                            */
                            ModeResetDataFlow();

                            retrieveText.Clear();
                            return;

                        case (retrieverMode.retrieveRuby):

                            ModeResetDataFlow();
                            prevRetrieveMode = retrieverMode.retrieveJS;
                            ModeSetRetJS();
                            retrieveText.Clear();
                            break;

                        case (retrieverMode.convertRubyToJS):

                            // Restore interface since the conversion mode disables it:
                            RestoreGuiConvToJS();

                            if (prevRetrieveMode == retrieverMode.retrieveRuby)
                            {
                                ModeResetDataFlow();
                                prevRetrieveMode = retrieverMode.retrieveJS;
                            }

                            ModeSetRetJS();
                            break;
                    }

                    break;

                case (retrieverMode.convertRubyToJS):

                    /*
                     * We would not be here if we were already in the mode [caller makes sure], so we must
                     * be in either 'Retrieve Ruby' or in 'Retrieve JS' mode.
                     *
                     * Disable most of the GUI since we are just converting what the user
                     * has copied to the clipboard, and setup for this mode.
                    */
                    DisableGuiConvToJS();

                    prevRetrieveMode = (retMode == retrieverMode.retrieveJS ? retMode : retrieverMode.retrieveRuby);

                    if (retMode == retrieverMode.retrieveRuby)
                    {
                        retrieveText.Clear();
                    }

                    retMode = retrieverMode.convertRubyToJS;

                    break;
            }
        }

        /*
         * Called when we need to remove prior dfObjects because they either contain
         * dataflow code written in the wrong language or a new document is being dataflowed
         * and the user does not want the 'Selection List' populated with another documents
         * selections - in this case they would be clicking on the retrieve mode menu
         * selection when that mode is already set to that selection.
         * 
         * Called From:
         * 
         *      ProcessMode()
        */
        public void ModeResetDataFlow()
        {
            indivDepMnemonicList.Clear();
            equSuccOffMnemonicList.Clear();
            equPredOffMnemonicList.Clear();
            indivMnemonicList.Clear();
            equSuccMnemonicList.Clear();
            equPredMnemonicList.Clear();
            equGenMnemonicList.Clear();
            equStateMnemonicList.Clear();
            entityMnemonicList.Clear();
            prepMnemonicList.Clear();
            fidMnemonicList.Clear();

            dfObject = null;
            dfObjectList.Clear();

            renumList.Items.Clear();
            mnemonicListBox.Items.Clear();
        }

        /*
         * This sets button and label text and sets the retMode for Retrieving Ruby.
         * 
         * Called From:
         * 
         *      ProcessMode()
        */
        public void ModeSetRetRuby()
        {
            retrieveButton.Text = "Retrieve Ruby Code";
            generateGlobalsButton.Text = "Generate $ Variables";
            modeLbl.Text = "Retrieve Ruby Code";

            retMode = retrieverMode.retrieveRuby;
        }

        /*
         * This sets button and label text and sets the retMode for Retrieving JavaScript.
         * 
         * Called From:
         * 
         *      ProcessMode()
        */
        public void ModeSetRetJS()
        {
            /*
             * The default for this is off, I think it would be annoying to have
             * the 'tmp' var declaration constantly present for every retrieval.
            */
            generateGlobalsButton.Text = "Gen Variables: OFF";

            retrieveButton.Text = "Retrieve JS Code";
            modeLbl.Text = "Retrieve JavaScript Code";

            retMode = retrieverMode.retrieveJS;
        }

        


        /*
         * Caller has determined that we are inserting over selected text, so we must make sure
         * that the selected text does not lie inside other text, else we end up with a mess.
         * 
         * Called From:
         *    RetrieveDataflow()
         *    condButton_Click()
         *
         * Inputs:
         *    selText: The selected text.
         *    
         * Returns:
         *    'true' if a good selection, else 'false' for a bad one, in which case we keep the
         *    text selected.
        */
        public bool GoodSelection(ref string selText)
        {
            bool result = true;
            string subStr;
            string appendNewLine;
            string retText;
            int startLine;
            int stopLine;
            int newLineCnt;

            /*
             * .SelectionStart is the index [zero based] into the .Text where the selection starts.
             * We use that and the .SelectionLength in determining the start and stop position index
             * values of the array version of .Text which is .Lines[].
             * 
             * Calculating startLine:
            */
            if (retrieveText.SelectionStart > 0)
            {
                subStr = retrieveText.Text.Substring(0, retrieveText.SelectionStart);
                startLine = CountNewLines(ref subStr);

                //Debug:
                //MessageBox.Show(string.Format("subStr:\n\n\"{0}\"", subStr), "Debug:", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
            else
            {
                startLine = 0;
            }

            // Calculating stopLine:
            newLineCnt = CountNewLines(ref selText);

            if (newLineCnt > 0)
            {
                stopLine = startLine + (Regex.IsMatch(selText, @"(\r\n?|\n)\z") ? newLineCnt - 1 : newLineCnt);
            }
            else
            {
                stopLine = startLine;
            }

            /*
             * Here we determine if the user has selected a block of text that does not lie
             * inside other text. We determine the first line of text from retrieveText using .Lines[]
             * that contains the start of the selected text and build up retText concatenating complete
             * Lines[] up to and including the last Lines[] where the selected text ends.
             * 
             * Note that we need to append a newline char to each pass through the loop because Lines[]
             * contains no newline chars.
             * 
             * The only time we do not append a newline is if the last line [in .Text] has no newline at
             * the end, this would only be for the very last line in Lines[]. Probably unlikely, but you
             * never know.
             * 
             * Basically, retText.Length will be greater than or equal to selText.Length. If their
             * lengths are not equal we will then compare the lengths after we trim away whitespace
             * chars from the right ends. We do this because people only 'see' non-whitespace chars,
             * and it would be reasonable for them to make a 'sloppy' selection and miss a newline
             * or a tab or a space at the end. They can clearly see, however, when they select from
             * the start of the text line and miss whitespace chars there, so we will hold them
             * accountable for that.
            */
            appendNewLine = "\n";
            retText = "";

            for (int index = startLine; index <= stopLine; index++)
            {
                // If we are on the last line....
                if (stopLine + 1 == retrieveText.Lines.Length)
                {
                    if (index == stopLine)
                    {
                        if (!Regex.IsMatch(retrieveText.Text, @"(\r\n?|\n)\z"))
                        {
                            appendNewLine = "";
                        }
                    }
                }

                retText += retrieveText.Lines[index] + appendNewLine;
            }

            if (selText != retText)
            {
                if (retText.TrimEnd() == selText.TrimEnd())
                {
                    /*
                     * They made a 'good' selection AFTER we trimmed whitespace chars from the ends of
                     * both strings. We will extend their selection to include the trailing whitespace
                     * chars. Trailing whitespace chars are easy for someone to miss. We will not be so
                     * kind with leading whitespace chars.
                    */
                    retrieveText.SelectionLength += retText.Length - selText.Length;

                    // I was doing this for condButton_Click().
                    // Re-read the selected text since we just added to the selection length:
                    //selText = retrieveText.SelectedText;
                }
                else
                {
                    // Not a good selection - keeping the selection.
                    ProcessErrors("Error!", 1, null, null);
                    retrieveText.Focus();
                    result = false;
                }
            }

            return result;
        }

        /*
         * This retrieves dataflow code, either Ruby or JavaScript as determined by the mode setting.
         * 
         * The actual dataflow code stored in the dfObject(s) is renumbered and truncated here.
         * We also perform conditional logic processing where we add the appropriate outer 'if ()' code
         * and insert the tabbed dataflow code followed by the appropriate closing conditional code.
         * 
         * The retrieved modified code is then concatenated into one string ready for insertion into
         * retrieveText. If text has been selected by the user we insert over the selection, else we
         * insert at the cursor location.
         * 
         * Called From:
         *    retrieveButton_Click()
         *    
         * NOTE: THIS ROUTINE IS ONLY CALLED WHEN RETRIEVING EITHER RUBY CODE OR JS CODE.
        */
        public void RetrieveDataflow()
        {
            bool doCond = false;
            bool noExtraLine = false;
            bool clearRenumList = false;
            bool resetSelList = false;

            /*
            * Common Block processing really throws a curve into this routine. DoCommonBlocks() will handle
            * Common Block [Individual dependents; Equifax Successor and Predecessor Officers] code retrieval.
            */
            bool doBlocks = false;
            int numBlocks = 1;
            Match match;

            List<string> codeList = new List<string>();
            List<string> dfStrList = new List<string>();

            string varName = "";
            string varStr = "";
            string selText = "";
            string retrieveStr = "";
            string blockType = "";
            string[] linesArray;

            int boxOffsetValue = 0;
            int loopCnt;

            // Not to be confused with global dfObjectList.
            List<DataFlow> dfObjList = new List<DataFlow>();

            CondExp condExp = null;

            // START: USED FOR CREATING STARTING AND STOPPING MNEMONIC LINES THAT WILL BE INSERTED INTO
            //        THE CODE SNIPPET FILE.

            //GetDataflowByMnemonic(xmlFile, mnemonic);
            //return;

            // STOP: USED FOR CREATING STARTING AND STOPPING MNEMONIC LINES THAT WILL BE INSERTED INTO
            //       THE CODE SNIPPET FILE.

            /*
            * We are retrieving [and possibly renumbering] from 1 of 3 sources:
            *
            * 1) We first look for highlighted selections in the 'Selection List' [mnemonicListBox] and
            * retrieve the associated dfObjects in the dfObjectList table and work with those. [the
            * dfObjectList is 1:1 with the 'Selection List']
            * 
            * 2) If not 1), we check if there is data in the 'Renumber and/or Truncate Selection Boxes'
            * list [renumList] and work with dfObject which would be set accordingly.
            * 
            * 3) If not 2) we then will set dfObject using the current value of the mnemonic [which has
            * been built via selections made in the various lists at the top of the screen] and there
            * would be no renumbering in that case.
            */
            if (mnemonicListBox.SelectedIndices.Count > 0)
            {
                for (int index = 0; index < mnemonicListBox.SelectedIndices.Count; index++)
                {
                    dfObjList.Add(dfObjectList[mnemonicListBox.SelectedIndices[index]]);
                }

                /*
                Flagging that we want to reset the selections in the 'Selection List'. Not actually doing
                it here as we may abort this routine due to an error, so we will check this flag just before
                we exit [at the end].
                */
                resetSelList = true;
            }
            else if (renumList.Items.Count > 0)
            {
                dfObjList.Add(dfObject);

                /*
                We want to clear this list since we retrieved from it, but only if we sucessfully
                performed the retrieval. This is done at the end of this routine.
                */
                clearRenumList = true;
            }
            else
            {
                // No sense in continuing if a null mnemonic.
                if (mnemonic == null)
                {
                    return;
                }

                /*
                We have to verify the mnemonic before we can proceed:
                */
                dfObject = GetDataflowByMnemonic(mnemonic, retMode == retrieverMode.retrieveJS);

                if (dfObject == null)
                {
                    retrieveText.Focus();
                    return;
                }

                dfObjList.Add(dfObject);

                /*
                Reset dfObject since we are retrieving from it. We do the same when retrieving from the
                'Renumber and/or Truncate Selection Boxes' list [at the end of this routine]. This repairs
                a 'logic leak' in the TK version. What would happen was, if you retrieve using the selection
                lists [top area] and then made a different selection and inserted into the 'Selection List',
                you would see the previous retrieved dataflow since dfObject was not null.
                */
                dfObject = null;
            }

            /*
            Check if we are processing common block [EX: Dependents] DataFLow, and set a flag if we are:
            [this is an all-or-none situation regarding common block data flow processing
            so it is sufficient to check just the first list element]
            */
            match = Regex.Match(dfObjList[0].Mnemonic, @"(DEPENDENTS|EQU SUCCESSOR\|OFFICERS|EQU PREDECESSOR\|OFFICERS)");
            if (match.Success)
            {
                /*
                 * First determine which type of common block we are processing, and set numBlock to the
                 * number of common blocks being processed and set the blockType.
                */
                switch (match.Groups[1].ToString())
                {
                    case ("DEPENDENTS"):
                        numBlocks = numDependents;
                        blockType = "dependent";
                        break;

                    case ("EQU SUCCESSOR|OFFICERS"):
                        numBlocks = numEquSuccOfficers;
                        blockType = "officer";
                        break;

                    case ("EQU PREDECESSOR|OFFICERS"):
                        numBlocks = numEquPredOfficers;
                        blockType = "officer";
                        break;
                }

                /*
                Put the cursor back in the retrieveText widget because the Box Offset entry
                or the manual Truncate field widget may have it.
                */
                retrieveText.Focus();

                /*
                 * We want to check 2 things first before we go any further with commom block processing:
                 * 
                 * 1) Check if numBlocks = 1 AND boxOffsetEntry > 0 in which case we warn them that
                 * boxOffsetEntry will be ignored.
                 * 
                 * 2) Check if numBlocks > 1 AND boxOffsetEntry = 0 in which case we popup an error
                 * message and abort.
                */
                if (boxOffsetEntry.Text.Length > 0)
                {
                    // Some defensive coding so TryParse is happy/reasonable:
                    if (Regex.IsMatch(boxOffsetEntry.Text, @"^\d{1,3}$"))
                    {
                        int.TryParse(boxOffsetEntry.Text, out boxOffsetValue);
                    }
                    else
                    {
                        ProcessErrors("Error!", 13, boxOffsetEntry.Text, "Box Offset");
                        boxOffsetEntry.Focus();
                        return;
                    }
                }

                if ((boxOffsetValue > 0) && (numBlocks == 1))
                {
                    ProcessErrors("Warning!", 9, blockType, null);
                    retrieveText.Focus();
                }
                else if ((boxOffsetValue == 0) && (numBlocks > 1))
                {
                    ProcessErrors("Error!", 8, blockType, null);
                    boxOffsetEntry.Focus();
                    return;
                }

                doBlocks = true;
            }

            /*
            Farming out the renumbering/truncation of the retrieved data flow code to ModifyDf( ).
            Each entry in dfStrList is the data flow code for 1 mnemonic, expressed as one string
            that has been renumbered and truncated by ModifyDF().         
            */
            dfStrList = ModifyDf(dfObjList);

            /*
            Now check if the retrieveText area has selected text [highlighted] in which case
            we overwrite that selected area, else we just insert at the next insertion location.
            */
            selText = retrieveText.SelectedText;

            if (selText.Length > 0)
            {
                if (!GoodSelection(ref selText))
                {
                    return;
                }

                // Because we are here we can do this [did not 'return;' above].
                noExtraLine = true;
            }

            /*
            We are now in the area where we convert the data flow code strings in the list dfStrList
            to one long string by concatenating the elements in dfStrList together. We do this
            not in one step, it is done as we move down through the code regions.

            In the Ruby version we were inserting into the retrieveText box line by line, but
            we can't do that here [well at first I thought that, maybe we could but got this far and
            ain't turning back...] , it goes in as one string via retrieveStr.
            */

            /*
             * Check for a conditional expression - Using Ruby for this explanation:
             * 
             * This is a multi-part process, we first insert the if (...) portion, then we insert
             * the mnemonic(s) [tabbed], then down below after inserting the mnemonic(s), we insert
             * an 'else' if applicable, then the terminating 'end'.
             * 
             * Conditional expressions are currently used for IND|SPOUSE and IND|SOLE PROPRIETOR
             * but I am processing everything for conditional expression anyways to facilitate future
             * additional conditional expression situations. Note that common block processing
             * [dependents are an example] falls into this 'not doing it now, maybe in the future'
             * scenario.
            */
            if (condList.SelectedIndices.Count > 0)
            {
                condExp = GetCond(profileTag, condList);

                // Adding some defensive coding here, THIS SHOULD NEVER HAPPEN, but.....
                if (condExp == null)
                {
                    ProcessErrors("Error! Tell Harry!!", 14, string.Format("GetCond(\"{0}\", condList)  \n\n  Which was called from:\n\n  retrieveButton_Click()", profileTag), null);
                    retrieveText.Focus();
                    return;
                }
                else
                {
                    doCond = true;

                    /*
                     * This is the very first assignment to retrieveStr so no concatenation. Just a FYI
                     * thingy.
                    */
                    retrieveStr = condExp.IfStr;
                }
            }

            if (doBlocks)
            {
                // Common block processing has some twists, so we divert over to DoCommonBlocks().
                retrieveStr += DoCommonBlocks(retrieveStr, numBlocks, noExtraLine, doCond, boxOffsetValue, dfStrList, blockType);
            }
            else
            {
                loopCnt = 0;

                for (int index = 0; index < dfStrList.Count; index++)
                {
                    loopCnt += 1;

                    if (doCond)
                    {
                        /*
                        We need to employ StringSplitOptions.RemoveEmptyEntries because, for example, "ABC\n"
                        split using StringSplitOptions.None would yield a 2 element array, with one element
                        being an empty string, which causes one more iteration thru the loop [does not hurt
                        anything text wise], but causes an extra newline to be re-attached down below [we
                        are splitting on newline chars so we need to add them back].
                        */
                        linesArray = dfStrList[index].Split(new string[] { "\r\n", "\n" }, StringSplitOptions.RemoveEmptyEntries);

                        foreach (string line in linesArray)
                        {
                            retrieveStr += string.Format("\t{0}\n", line);
                        }
                    }
                    else
                    {
                        retrieveStr += dfStrList[index];
                    }

                    if ((dfStrList.Count > 1) && (loopCnt < dfStrList.Count))
                    {
                        retrieveStr += "\n";
                    }
                }

                if (!doCond)
                {
                    retrieveStr += (noExtraLine ? "" : "\n");
                }
            }

            if (doCond)
            {
                /*
                 * Now the rest of the conditional processing:
                 * 
                 * The 'else' part of the conditional expression if aplicable.
                */
                if (condExp.ElseStr != null)
                {
                    retrieveStr += condExp.ElseStr;
                }

                // The 'end' of the conditional expression:
                retrieveStr += string.Format("{0}\n{1}", (retMode == retrieverMode.retrieveJS ? "}" : "end"), (noExtraLine ? "" : "\n"));

                // Now clear the conditional selection:
                condList.ClearSelected();
            }

            /*
             * Now we are ready to actually 'insert' the retrieved dataflow code that MAY have been
             * adjusted for applicable renumbering/truncating/conditional processing. As strange as this
             * looks, it works. The key is that when text has been selected, SelectionLength is > 0, which
             * means that the selected text will be replaced. When SelectedLength is = 0, it means to
             * insert the text at the current cursor location.
             * 
             * NOTE: Since we changed to StringSplitOptions.RemoveEmptyEntries for the .Split on the
             * newline chars, we fixed one problem (extra newlines added) and created another: when we read
             * from the code snippet file and we read a blank line, that is, a line that has just a newline
             * char, we lose that line due to the StringSplitOptions.RemoveEmptyEntries, so in order to
             * keep those blank lines I added a space char followed by a newline char in
             * GetDataflowByMnemonic(). Here I will replace instances of " \n" with "\n" before I write the
             * retrieved text to the retrieveText text box.
            */
            Regex regex = new Regex(@" \n");
            retrieveStr = regex.Replace(retrieveStr, "\n");

            /*
             * If we are retrieving JavaScript we need to check if we are to also generate a var list
             * which we need to output before the actual dataflow code. dfStrList holds all the dataflow
             * with each list element being one dataflow snippet, all on 'one' line, so we split each of
             * these lines on the newline and then inspect each for a variable assignment.
            */
            if (retMode == retrieverMode.retrieveJS)
            {
                if (generateGlobalsButton.Text == "Gen Variables: ON")
                {
                    foreach (string dfStr in dfStrList)
                    {
                        codeList = dfStr.Split(new string[] { "\r\n", "\n" }, StringSplitOptions.RemoveEmptyEntries).ToList();

                        foreach (string codeLine in codeList)
                        {
                            /*
                             * The (?i) says to ignore case. Did this because I was stubbing in Ruby
                             * code that was upper cased to simulate the JS code. Leaving this in since
                             * some vars in the JS use upper cased chars. [camel casing I guess].
                             */
                            match = Regex.Match(codeLine, @"(?i)([a-z][a-z0-9_]+)\s+=\s+.+");

                            // Original code for var indentification.
                            //match = Regex.Match(codeLine, @"(?i)^\s*([a-z][a-z0-9_]+)\s*=\s*.+");

                            if (match.Success)
                            {
                                varName = string.Format("var {0} = '';{1}", match.Groups[1].Value, Environment.NewLine);

                                if (!varStr.Contains(varName))
                                {
                                    varStr += varName;
                                }
                            }
                        }
                    }
                }
            }

            retrieveText.SelectedText = (varStr.Length > 0 ? varStr + Environment.NewLine : "") + retrieveStr;

            // Making sure they see what they just inserted:
            retrieveText.ScrollToCaret();
            retrieveText.Focus();

            /*
             * We retrieved the code, now we need to clear the 'Renumber and/or Truncate Selection' Boxes
             * list IF we retrieved from it so we do not confuse the user, you know how those users are.
             * They could retrieve from the list [instead of from the 'Selection List' or from the lists
             * at the top of the screen] and then select from the top lists and attempt to retrieve from
             * that area but get the 'Renumber and/or Truncate Selection' selection instead because it was
             * used instead - we retrieve from the 'Renumber and/or Truncate Selection' list if nothing is
             * selected in the 'Selection List'.
            */
            if (clearRenumList)
            {
                dfObject = null;
                renumList.Items.Clear();

                // De-activate the Manual Truncate entry field and clear it out:
                manTruncEntry.Clear();
                renumTruncPanel.Enabled = false;

                /*
                 * Put the cursor back in the retrieveText box because the Manual Truncate
                 * entry widgy thingy may have had it if they used it.
                */
                retrieveText.Focus();
            }

            if (resetSelList)
            {
                mnemonicListBox.ClearSelected();
            }
        }

        public void ProcessRubyConversion()
        {
            string selText;

            if (Clipboard.ContainsText(TextDataFormat.Text))
            {
                convertRuby = new ConvertRuby(ref xmlDocument, generateGlobalsButton.Text == "Gen Variables: ON");

                convertRuby.DoConversion();

                if (!String.IsNullOrEmpty(convertRuby.convertStr))
                {
                    /*
                     * convertRuby.convertStr has also been copied to the clipboard.
                     * 
                     * If text has been selected by the user we insert over the selection, else we
                     * insert at the cursor location.
                     */
                    selText = retrieveText.SelectedText;

                    if (selText.Length > 0)
                    {
                        if (!GoodSelection(ref selText))
                        {
                            return;
                        }
                    }

                    retrieveText.SelectedText = convertRuby.convertStr;

                    // Making sure they see what they just inserted:
                    retrieveText.ScrollToCaret();
                    retrieveText.Focus();
                }
            }
        }

    }
}